import logging
from types import MappingProxyType
from typing import Any

import fitz
from pyspark import keyword_only

from scaledp.enums import ImageType
from scaledp.image.DataToImage import DataToImage
from scaledp.params import HasResolution, Param, Params, TypeConverters
from scaledp.schemas.Image import Image


class PdfDataToSingleImage(DataToImage, HasResolution):
    pageNumber = Param(
        Params._dummy(),
        "pageNumber",
        "Page number to convert to image",
        typeConverter=TypeConverters.toInt,
    )

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "content",
            "outputCol": "image",
            "pathCol": "path",
            "keepInputData": False,
            "imageType": ImageType.FILE,
            "resolution": 300,
            "pageNumber": 0,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfDataToSingleImage, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def process(self, input, path, resolution=0):
        logging.info("Run Pdf to Image")
        try:
            doc = fitz.open("pdf", input)
            if len(doc) == 0:
                raise ValueError("Empty PDF document.")

            pix = doc[self.getPageNumber()].get_pixmap(
                matrix=fitz.Identity,
                dpi=self.getResolution(),
                colorspace=fitz.csRGB,
                clip=None,
                alpha=False,
                annots=True,
            )

            return Image.from_binary(
                pix.pil_tobytes("png"),
                path,
                self.getImageType(),
                resolution=self.getResolution(),
                width=pix.width,
                height=pix.height,
            )

        except Exception:
            return Image(
                path,
                exception="Error in extration of image from pdf document",
            )

    def getPageNumber(self):
        return self.getOrDefault(self.pageNumber)

    def setPageNumber(self, value):
        return self._set(pageNumber=value)



from types import MappingProxyType
from typing import Any

import numpy
from pyspark import keyword_only

from scaledp.models.detectors.craft.models import Craftnet, Refinenet
from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice, Param, Params, TypeConverters
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class CraftTextDetector(BaseDetector, HasDevice, HasBatchSize):
    _craft_net = None
    _refine_net = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.7,
            "textThreshold": 0.4,
            "linkThreshold": 0.4,
            "sizeThreshold": -1,
            "width": 1280,
            "withRefiner": False,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
        },
    )

    textThreshold = Param(
        Params._dummy(),
        "textThreshold",
        "Threshold for text region score",
        typeConverter=TypeConverters.toFloat,
    )

    linkThreshold = Param(
        Params._dummy(),
        "linkThreshold",
        "Threshold for link affinity score",
        typeConverter=TypeConverters.toFloat,
    )

    sizeThreshold = Param(
        Params._dummy(),
        "sizeThreshold",
        "Threshold for height of detected regions",
        typeConverter=TypeConverters.toInt,
    )

    width = Param(
        Params._dummy(),
        "width",
        "Width for image resizing",
        typeConverter=TypeConverters.toInt,
    )

    withRefiner = Param(
        Params._dummy(),
        "withRefiner",
        "Enable refiner network postprocessing",
        typeConverter=TypeConverters.toBoolean,
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(CraftTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._craft_net and cls._refine_net:
            return cls._craft_net, cls._refine_net

        # from craft_text_detector import load_craftnet_model, load_refinenet_model
        #
        # device = "cuda" if int(params["device"]) == Device.CUDA.value else "cpu"
        # use_cuda = device == "cuda"
        #
        # craft_net = load_craftnet_model(
        #     cuda=use_cuda,
        # )
        #
        # refine_net = None
        # if params.get("withRefiner"):
        #     refine_net = load_refinenet_model(
        #         cuda=use_cuda,
        #     )

        craft_net = Craftnet()
        # load refinernet if required
        refine_net = None
        if params.get("withRefiner"):
            refine_net = Refinenet()

        cls._craft_net = craft_net
        cls._refine_net = refine_net
        return craft_net, refine_net

    @classmethod
    def call_detector(cls, images, params):
        import cv2
        from crafter import craft_utils, image_utils



        craft_net, refine_net = cls.get_model(params)
        use_cuda = int(params["device"]) == Device.CUDA.value
        results = []

        for image, image_path in images:
            try:
                # Convert PIL to OpenCV format
                #img_cv = numpy.array(image)[:, :, ::-1]

                img_cv = image_utils.read_image(numpy.array(image)[:, :, ::-1])

                # Resize
                img_resized, target_ratio, size_heatmap = image_utils.resize_aspect_ratio(
                    img_cv,
                    params["width"],
                    interpolation=cv2.INTER_LINEAR,
                )
                ratio_h = ratio_w = 1 / target_ratio


                # Preprocess
                x = image_utils.normalizeMeanVariance(img_resized)
                x = numpy.transpose(x, (2, 0, 1))  # [h, w, c] to [c, h, w]
                x = numpy.expand_dims(x, 0)  # [c, h, w] to [b, c, h, w]

                # Forward pass
                # with torch_utils.no_grad():
                #     y, feature = craft_net(x)
                y, feature = craft_net(x)

                # score_text = y[0, :, :, 0] #.cpu().data.numpy()
                # score_link = y[0, :, :, 1] #.cpu().data.numpy()

                score_text = y[0, 0, :, :]
                score_link = y[0, 1, :, :]

                #print(score_text)

                # Refine if enabled
                # if refine_net is not None:
                #     with torch_utils.no_grad():
                #         y_refiner = refine_net(y, feature)
                #     score_link = y_refiner[0, :, :, 0].cpu().data.numpy()

                if refine_net is not None:
                    y_refiner = refine_net(y, feature)
                    score_link = y_refiner[0, 0, :, :]

                # Post-process
                boxes, _ = craft_utils.getDetBoxes(
                    score_text,
                    score_link,
                    params["scoreThreshold"],
                    params["linkThreshold"],
                    params["textThreshold"],
                    poly=False,
                )

                # Adjust coordinates
                boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)


                # Convert to Box objects
                box_objects = [Box.from_polygon(box) for box in boxes]

                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=box_objects,
                        exception="",
                    ),
                )

            except Exception as e:
                raise e
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=[],
                        exception=f"CraftTextDetector error: {e!s}",
                    ),
                )

        # if use_cuda:
        #     torch.cuda.empty_cache()

        return results

    def setTextThreshold(self, value):
        return self._set(textThreshold=value)

    def setLinkThreshold(self, value):
        return self._set(linkThreshold=value)

    def setSizeThreshold(self, value):
        return self._set(sizeThreshold=value)

    def setWidth(self, value):
        return self._set(width=value)

    def setWithRefiner(self, value):
        return self._set(withRefiner=value)



# from pyspark.ml.param.shared import HasInputCol, HasOutputCol
# from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable, JavaMLReadable, JavaMLWritable
# from pyspark.ml.wrapper import JavaTransformer
# from pyspark.ml.param import Param, Params, TypeConverters
# from pyspark.sql.types import *
# from pyspark import keyword_only
# from pathlib import Path
#
# import json
# import time
# import logging
# import traceback
#
# import pyspark.sql.functions as f
# import torch
# from pyspark.ml import Transformer
# from cleverdoc.schemas import Box, Image
# from cleverdoc.params import HasKeepInput
#
#
#
# class TextDetector(Transformer, HasInputCol, HasOutputCol,
#                           HasKeepInput, DefaultParamsReadable, DefaultParamsWritable, JavaMLReadable, JavaMLWritable):
#     name = "TextDetector"
#
#     scoreThreshold = Param(Params._dummy(), "scoreThreshold",
#                            "Score threshold above which detection is considered as reliable",
#                            typeConverter=TypeConverters.toFloat)
#     textThreshold = Param(Params._dummy(), "textThreshold",
#                           "Threshold for the region(text) score. The region score represents \
#                           the probability that the given pixel is the center of the character",
#                           typeConverter=TypeConverters.toFloat)
#     linkThreshold = Param(Params._dummy(), "linkThreshold",
#                           "Threshold for the the link(affinity) score. The link score represents \
#                           the center probability of the space between adjacent characters",
#                           typeConverter=TypeConverters.toFloat)
#     sizeThreshold = Param(Params._dummy(), "sizeThreshold",
#                           "Threshold for height of the detected regions",
#                           typeConverter=TypeConverters.toInt)
#     withRefiner = Param(Params._dummy(), "withRefiner",
#                         "Enable to run refiner net as postprocessing step.",
#                         typeConverter=TypeConverters.toBoolean)
#     width = Param(Params._dummy(), "width",
#                   "Width of the desired input image. Image will be resized to this width.",
#                   typeConverter=TypeConverters.toInt)
#
#     mergeIntersects = Param(Params._dummy(), "mergeIntersects",
#                             "Merge intersects boxes.",
#                             typeConverter=TypeConverters.toBoolean)
#
#     useGPU = Param(Params._dummy(), "useGPU",
#                    "Try to use the GPU if available.",
#                    typeConverter=TypeConverters.toBoolean)
#
#     usePandasUdf = Param(Params._dummy(), "usePandasUdf",
#                          "Use Pandas Udf.",
#                          typeConverter=TypeConverters.toBoolean)
#
#     @keyword_only
#     def __init__(self):
#         super(TextDetector, self).__init__()
#
#         self._setDefault(scoreThreshold=0.7)
#         self._setDefault(textThreshold=0.4)
#         self._setDefault(linkThreshold=0.4)
#         self._setDefault(width=1280)
#         self._setDefault(sizeThreshold=-1)
#         self._setDefault(mergeIntersects=True)
#         self._setDefault(forceProcessing=True)
#         self._setDefault(keepInput=True)
#         self._setDefault(useGPU=False)
#         self._setDefault(usePandasUdf=False)
#
#
#     def setInputCol(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(inputCol=value)
#
#     def setOutputCol(self, value):
#         """
#         Sets the value of :py:attr:`outputCol`.
#         """
#         return self._set(outputCol=value)
#
#     @staticmethod
#     def get_prediction(
#             image,
#             craft_net,
#             refine_net=None,
#             text_threshold: float = 0.7,
#             link_threshold: float = 0.4,
#             low_text: float = 0.4,
#             cuda: bool = False,
#             long_size: int = 1280,
#             poly: bool = True,
#     ):
#
#         import cv2
#
#         from craft_text_detector import craft_utils
#         from craft_text_detector import image_utils
#         from craft_text_detector import torch_utils
#         # pylint: disable=pointless-string-statement
#         """
#         Arguments:scoreThreshold
#             image: path to the image to be processed or numpy array or PIL image
#             output_dir: path to the results to be exported
#             craft_net: craft net model
#             refine_net: refine net model
#             text_threshold: text confidence threshold
#             link_threshold: link confidence threshold
#             low_text: text low-bound score
#             cuda: Use cuda for inference
#             canvas_size: image size for inference
#             long_size: desired longest image size for inference
#             poly: enable polygon type
#         Output:
#             {"masks": lists of predicted masks 2d as bool array,
#              "boxes": list of coords of points of predicted boxes,
#              "boxes_as_ratios": list of coords of points of predicted boxes as ratios of image size,
#              "polys_as_ratios": list of coords of points of predicted polys as ratios of image size,
#              "heatmaps": visualizations of the detected characters/links,
#              "times": elapsed times of the sub modules, in seconds}
#         """
#         t0 = time.time()
#
#         # read/convert image
#         image = image_utils.read_image(image)
#
#         # resize
#         img_resized, target_ratio, size_heatmap = image_utils.resize_aspect_ratio(
#             image, long_size, interpolation=cv2.INTER_LINEAR
#         )
#         ratio_h = ratio_w = 1 / target_ratio
#         resize_time = time.time() - t0
#         t0 = time.time()
#
#         # preprocessing
#         x = image_utils.normalizeMeanVariance(img_resized)
#         x = torch_utils.from_numpy(x).permute(2, 0, 1)  # [h, w, c] to [c, h, w]
#         x = torch_utils.Variable(x.unsqueeze(0))  # [c, h, w] to [b, c, h, w]
#         if cuda:
#             x = x.cuda()
#         preprocessing_time = time.time() - t0
#         t0 = time.time()
#
#         # forward pass
#         with torch_utils.no_grad():
#             y, feature = craft_net(x)
#         craftnet_time = time.time() - t0
#         t0 = time.time()
#
#         # make score and link map
#         score_text = y[0, :, :, 0].cpu().data.numpy()
#         score_link = y[0, :, :, 1].cpu().data.numpy()
#
#         # refine link
#         if refine_net is not None:
#             with torch_utils.no_grad():
#                 y_refiner = refine_net(y, feature)
#             score_link = y_refiner[0, :, :, 0].cpu().data.numpy()
#         refinenet_time = time.time() - t0
#         t0 = time.time()
#
#         # Post-processing
#         boxes, polys = craft_utils.getDetBoxes(
#             score_text, score_link, text_threshold, link_threshold, low_text, poly
#         )
#
#         # coordinate adjustment
#         boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)
#
#         postprocess_time = time.time() - t0
#
#         times = {
#             "resize_time": resize_time,
#             "preprocessing_time": preprocessing_time,
#             "craftnet_time": craftnet_time,
#             "refinenet_time": refinenet_time,
#             "postprocess_time": postprocess_time,
#         }
#
#         return {
#             "boxes": boxes,
#             "times": times,
#         }
#
#     @staticmethod
#     def udf(image, model_path, params, exception):
#         logging.info("Run ImageTextDetectorV2")
#
#         try:
#
#             params = json.loads(params)
#             use_cuda = torch.cuda.is_available() and params.get("useGPU")
#
#             from craft_text_detector import (
#                 load_craftnet_model,
#                 load_refinenet_model,
#                 # get_prediction,
#                 empty_cuda_cache
#             )
#             # read image
#             image = to_opencv_image(image)
#
#             # load models
#             refine_net = None
#             if params.get("withRefiner"):
#                 refine_net = load_refinenet_model(cuda=use_cuda,
#                                                   weight_path=str(Path(model_path, "craft_refiner_CTW1500.pth")))
#             craft_net = load_craftnet_model(cuda=use_cuda, weight_path=str(Path(model_path, "craft_mlt_25k.pth")))
#
#             long_size = int(params.get("width") if params.get("width") != 0 else 1280)
#             # perform prediction
#             prediction_result = ImageTextDetectorV2.get_prediction(
#                 image=image,
#                 craft_net=craft_net,
#                 refine_net=refine_net,
#                 text_threshold=params.get("scoreThreshold"),
#                 link_threshold=params.get("linkThreshold"),
#                 low_text=params.get("textThreshold"),
#                 cuda=use_cuda,
#                 long_size=long_size,
#                 poly=False
#             )
#
#             rects = [Rectangle.from_box(box) for box in prediction_result["boxes"]]
#             character_height = get_size(rects, lambda x: x.get_height())
#
#             size_threshold = params.get("sizeThreshold")
#             if params.get("sizeThreshold") == -1:
#                 size_threshold = int(character_height / 2)
#             line_tolerance = int(character_height / 2)
#
#             rects = list(filter(lambda x: x.get_height() > size_threshold, rects))
#
#             if params.get("mergeIntersects"):
#                 rects = Rectangle.merge_intersects(rects, line_tolerance)
#
#             regions = [rect.to_region() for rect in rects]
#         except Exception as error:
#             logging.exception(error)
#             regions = []
#             exception = "ImageTextDetectorV2: Error during text detection: " + str(traceback.format_exc()) + str(
#                 exception)
#         finally:
#             if use_cuda:
#                 empty_cuda_cache()
#
#         return (regions, exception)
#
#     @staticmethod
#     def pandas_udf(images, model_path, params, exception):
#         logging.info("Run ImageTextDetectorV2")
#
#         params = json.loads(params[0])
#         use_cuda = torch.cuda.is_available() and params.get("useGPU")
#
#         exception = exception[0]
#
#         import pandas as pd
#
#         from craft_text_detector import (
#             load_craftnet_model,
#             load_refinenet_model,
#             empty_cuda_cache
#         )
#
#         # load models
#         refine_net = None
#         if params.get("withRefiner"):
#             refine_net = load_refinenet_model(cuda=use_cuda,
#                                               weight_path=str(Path(model_path[0], "craft_refiner_CTW1500.pth")))
#         craft_net = load_craftnet_model(cuda=use_cuda, weight_path=str(Path(model_path[0], "craft_mlt_25k.pth")))
#
#         long_size = int(params.get("width") if params.get("width") != 0 else 1280)
#
#         res = []
#         for i, image in images.iterrows():
#             logging.info(f"ImageTextDetectorV2: process image {i}")
#             try:
#                 # read image
#                 image = to_opencv_image(Image(**image))
#                 prediction_result = ImageTextDetectorV2.get_prediction(
#                     image=image,
#                     craft_net=craft_net,
#                     refine_net=refine_net,
#                     text_threshold=params.get("scoreThreshold"),
#                     link_threshold=params.get("linkThreshold"),
#                     low_text=params.get("textThreshold"),
#                     cuda=use_cuda,
#                     long_size=long_size,
#                     poly=False
#                 )
#
#                 rects = [Rectangle.from_box(box) for box in prediction_result["boxes"]]
#                 character_height = get_size(rects, lambda x: x.get_height())
#
#                 size_threshold = params.get("sizeThreshold")
#                 if params.get("sizeThreshold") == -1:
#                     size_threshold = int(character_height / 2)
#                 line_tolerance = int(character_height / 2)
#
#                 rects = list(filter(lambda x: x.get_height() > size_threshold, rects))
#
#                 if params.get("mergeIntersects"):
#                     rects = Rectangle.merge_intersects(rects, line_tolerance)
#
#                 regions = [rect.to_region() for rect in rects]
#
#             except Exception as error:
#                 logging.exception(error)
#                 regions = []
#                 exception = "ImageToTextV2: Error during text detection: " + traceback.format_exc() + exception
#             res.append(json.dumps({params['outputCol']: regions, "exception": exception}))
#         if use_cuda:
#             empty_cuda_cache()
#
#         return pd.Series(res)
#
#     def _transform(self, dataset):
#
#         out_col = self.getOutputCol()
#
#         if self.getInputCol() not in dataset.columns:
#             uid_ = self.uid
#             inp_col = self.getInputCol()
#             raise ValueError(f"""Missing input column in {uid_}: Column '{inp_col}' is not present.
#         Make sure such transformer exist in your pipeline,
#         with the right output names.""")
#
#         image_col = self.getInputCol()
#
#         params = json.dumps({k.name: v for k, v in self.extractParamMap().items()})
#
#         output_schema = StructType([StructField(out_col, CoordinateSchema, True),
#                                     StructField('exception', StringType(), True)])
#
#         if "exception" not in dataset.columns:
#             dataset = dataset.withColumn("exception", f.lit(""))
#
#         if self.getOrDefault(self.usePandasUdf):
#             sel_cols_input = [f.from_json(f.pandas_udf(ImageTextDetectorV2.pandas_udf, StringType())(image_col,
#                                                                                                      f.lit(
#                                                                                                          self.getModelPath()),
#                                                                                                      f.lit(params),
#                                                                                                      "exception"),
#                                           output_schema).alias("tmp_result")] + \
#                              [f.col(x) for x in dataset.columns if
#                               x not in (image_col, "exception")]
#         else:
#             sel_cols_input = [self.exploding_wrap(f.udf(ImageTextDetectorV2.udf, output_schema))(image_col,
#                                                                                                  f.lit(
#                                                                                                      self.getModelPath()),
#                                                                                                  f.lit(params),
#                                                                                                  "exception").alias(
#                 "tmp_result")] + \
#                              [f.col(x) for x in dataset.columns if
#                               x not in (image_col, "exception")]
#
#         sel_cols = [f.col("tmp_result.*")] + \
#                    [f.col(x) for x in dataset.columns if x not in (image_col, "exception")]
#
#         if self.getOrDefault(self.keepInput):
#             sel_cols_input.append(f.col(image_col))
#             sel_cols.append(f.col(image_col))
#
#         result = dataset.select(*sel_cols_input) \
#             .select(*sel_cols)
#         if not self.getOrDefault(self.keepInput):
#             result = result.drop(image_col)
#         return result
#
#     def setTextThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(textThreshold=value)
#
#     def setLinkThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(linkThreshold=value)
#
#     def setScoreThreshold(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(scoreThreshold=value)
#
#     def setWithRefiner(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(withRefiner=value)
#
#     def setUseGPU(self, value):
#         """
#         Sets the value of :py:attr:`useGPU`.
#         """
#         return self._set(useGPU=value)
#
#     def setSizeThreshold(self, value):
#         """
#         Sets the value of :py:attr:`sizeThreshold`.
#         """
#         return self._set(sizeThreshold=value)
#
#     def setWidth(self, value):
#         """
#         Sets the value of :py:attr:`inputCol`.
#         """
#         return self._set(width=value)
#
#     def setMergeIntersects(self, value):
#         """
#         Sets the value of :py:attr:`mergeIntersects`.
#         """
#         return self._set(mergeIntersects=value)
#
#     def setUsePandasUdf(self, value):
#         """
#         Sets the value of :py:attr:`spaceWidth`.
#         """
#         return self._set(usePandasUdf=value)



## Frome the https://github.com/xdanny/pyspark_types/blob/main/pyspark_types/dataclass.py
## with some changes

import datetime
from dataclasses import fields, is_dataclass
from typing import Dict, Type, Union, get_type_hints

from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DataType,
    DateType,
    DecimalType,
    DoubleType,
    IntegerType,
    LongType,
    MapType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from scaledp.utils.auxiliary import BinaryT, BoundDecimal, ByteT, LongT, ShortT

type_mapping = {
    str: StringType,
    int: IntegerType,
    LongT: LongType,
    ShortT: ShortType,
    ByteT: ByteType,
    float: DoubleType,
    datetime.datetime: TimestampType,
    datetime.date: DateType,
    bool: BooleanType,
    BinaryT: BinaryType,
}


def register_type(new_type: Type, spark_type: DataType) -> None:
    """Register type mapping for a new type."""
    type_mapping[new_type] = spark_type


def map_dataclass_to_struct(dataclass_type: Type) -> StructType:
    """
    Map a Python data class to a PySpark struct.

    :param dataclass_type: The Python data class to be mapped.
    :return: A PySpark struct that corresponds to the data class.
    """
    fields_list = []
    hints = get_type_hints(dataclass_type)

    for field in fields(dataclass_type):
        field_name = field.name
        field_type = field.type

        if is_dataclass(field_type):
            # Recursively map nested data classes to PySpark structs
            sub_struct = map_dataclass_to_struct(field_type)
            nullable = is_field_nullable(field_name, hints)
            fields_list.append(StructField(field_name, sub_struct, nullable))
        elif hasattr(field_type, "__origin__") and field_type.__origin__ is list:
            # Handle lists of elements
            elem_type = field_type.__args__[0]
            if is_dataclass(elem_type):
                # Handle lists of data classes
                sub_struct = map_dataclass_to_struct(elem_type)
                nullable = is_field_nullable(field_name, hints)
                fields_list.append(
                    StructField(field_name, ArrayType(sub_struct), nullable),
                )
            else:
                # Handle lists of primitive types and dicts
                spark_type = get_spark_type(elem_type)
                nullable = is_field_nullable(field_name, hints)
                if spark_type == MapType(StringType(), StringType()):
                    # Special case for dictionaries with any value type
                    fields_list.append(StructField(field_name, spark_type, nullable))
                else:
                    fields_list.append(
                        StructField(field_name, ArrayType(spark_type), nullable),
                    )
        elif hasattr(field_type, "__origin__") and field_type.__origin__ is dict:
            # Handle dictionaries
            key_type, value_type = field_type.__args__
            if is_dataclass(value_type):
                sub_struct = map_dataclass_to_struct(value_type)
                nullable = is_field_nullable(field_name, hints)
                fields_list.append(
                    StructField(
                        field_name,
                        MapType(get_spark_type(key_type), sub_struct),
                        nullable,
                    ),
                )
            else:
                spark_type = get_spark_type(value_type)
                nullable = is_field_nullable(field_name, hints)
                fields_list.append(
                    StructField(
                        field_name,
                        MapType(get_spark_type(key_type), spark_type),
                        nullable,
                    ),
                )
        else:
            # Handle primitive types and BoundDecimal custom type
            spark_type = get_spark_type(field_type)
            nullable = is_field_nullable(field_name, hints)
            fields_list.append(StructField(field_name, spark_type, nullable))

    return StructType(fields_list)


def get_spark_type(
    py_type: Type,
    type_mapping: Dict[Type, DataType] = type_mapping,
) -> DataType:
    """Creates a mapping from a python type to a pyspark data type."""

    # Check if the type exists in the mapping
    if py_type in type_mapping:
        # If it's a function, call it (e.g., for Box.getSchema())
        return (
            type_mapping[py_type]()
            if callable(type_mapping[py_type])
            else type_mapping[py_type]()
        )

    # Handle BoundDecimal separately as it needs specific attributes
    if isinstance(py_type, type) and issubclass(py_type, BoundDecimal):
        return DecimalType(precision=py_type.precision, scale=py_type.scale)

    # Handle Optional types
    if is_optional_type(py_type):
        elem_type = py_type.__args__[0]
        return get_spark_type(elem_type)

    raise Exception(f"Type {py_type} is not supported by PySpark")


def is_field_nullable(field_name: str, hints: dict) -> bool:
    """
    Returns True if the given field name is nullable, based on the
    type hint for the field in the given hints dictionary.
    """

    if field_name not in hints:
        return True
    field_type = hints[field_name]
    return is_optional_type(field_type)


def apply_nullability(dtype: DataType, is_nullable: bool) -> DataType:
    """Returns a new PySpark DataType with the nullable flag set to the given value."""
    if is_nullable:
        if isinstance(dtype, StructType):
            # Wrap the nullable field in a struct with a single field
            return StructType([StructField("value", dtype, True)])
        if hasattr(dtype, "add_nullable"):
            return dtype.add_nullable()
        raise TypeError(f"Type {dtype} does not support nullability")
    return dtype


def is_optional_type(py_type: Type) -> bool:
    """Returns True if the given type is an Optional type."""
    if hasattr(py_type, "__origin__") and py_type.__origin__ is Union:
        args = py_type.__args__
        if len(args) == 2 and args[1] is type(None):
            return True
    return False



import json

from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import lit, pandas_udf, udf
from pyspark.sql.types import (
    ArrayType,
    DoubleType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

from scaledp.params import (
    HasBatchSize,
    HasColumnValidator,
    HasDefaultEnum,
    HasDevice,
    HasInputCol,
    HasKeepInputData,
    HasModel,
    HasNumPartitions,
    HasOutputCol,
    HasPageCol,
    HasPartitionMap,
    HasPathCol,
    HasScoreThreshold,
    HasWhiteList,
)
from scaledp.schemas.NerOutput import NerOutput


class BaseNer(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasWhiteList,
    HasDevice,
    HasModel,
    HasPathCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasNumPartitions,
    HasScoreThreshold,
    HasBatchSize,
    HasPageCol,
    HasColumnValidator,
    HasDefaultEnum,
    HasPartitionMap,
):

    def outputSchema(self):
        return StructType(
            [
                StructField("path", StringType(), True),
                StructField(
                    "entities",
                    ArrayType(
                        StructType(
                            [
                                StructField("entity_group", StringType(), False),
                                StructField("score", DoubleType(), False),
                                StructField("word", StringType(), False),
                                StructField("start", IntegerType(), False),
                                StructField("end", IntegerType(), False),
                                StructField(
                                    "boxes",
                                    ArrayType(
                                        StructType(
                                            [
                                                StructField(
                                                    "text",
                                                    StringType(),
                                                    False,
                                                ),
                                                StructField(
                                                    "score",
                                                    DoubleType(),
                                                    False,
                                                ),
                                                StructField("x", IntegerType(), False),
                                                StructField("y", IntegerType(), False),
                                                StructField(
                                                    "width",
                                                    IntegerType(),
                                                    False,
                                                ),
                                                StructField(
                                                    "height",
                                                    IntegerType(),
                                                    False,
                                                ),
                                            ],
                                        ),
                                        True,
                                    ),
                                    False,
                                ),
                            ],
                        ),
                        True,
                    ),
                    True,
                ),
                StructField("exception", StringType(), True),
            ],
        )

    def get_params(self):
        return json.dumps({k.name: v for k, v in self.extractParamMap().items()})

    def _transform(self, dataset):
        params = self.get_params()
        out_col = self.getOutputCol()
        in_col = self._validate(self.getInputCol(), dataset)

        if not self.getPartitionMap():
            result = dataset.withColumn(
                out_col,
                udf(self.transform_udf, NerOutput.get_schema())(in_col),
            )
        else:
            if self.getNumPartitions() > 0:
                if self.getPageCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPageCol())
                elif self.getPathCol() in dataset.columns:
                    dataset = dataset.repartition(self.getPathCol())
                dataset = dataset.coalesce(self.getNumPartitions())
            result = dataset.withColumn(
                out_col,
                pandas_udf(self.transform_udf_pandas, self.outputSchema())(
                    in_col,
                    lit(params),
                ),
            )

        if not self.getKeepInputData():
            result = result.drop(in_col)
        return result



import logging
import traceback
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import udf

from scaledp.params import (
    AutoParamsMeta,
    HasColor,
    HasColumnValidator,
    HasDefaultEnum,
    HasImageType,
    HasInputCols,
    HasKeepInputData,
    HasNumPartitions,
    HasOutputCol,
    HasPageCol,
    HasPropagateExc,
    Param,
    Params,
    TypeConverters,
)
from scaledp.schemas.Box import Box
from scaledp.schemas.Image import Image

from ..enums import ImageType


class ImageCropError(Exception):
    pass


class ImageCropBoxes(
    Transformer,
    HasInputCols,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColor,
    HasNumPartitions,
    HasColumnValidator,
    HasDefaultEnum,
    HasPropagateExc,
    metaclass=AutoParamsMeta,
):
    """Crop image by bounding boxes."""

    padding = Param(
        Params._dummy(),
        "padding",
        "Padding.",
        typeConverter=TypeConverters.toInt,
    )
    noCrop = Param(
        Params._dummy(),
        "noCrop",
        "Does not Crop if boxes is empty.",
        typeConverter=TypeConverters.toBoolean,
    )

    defaultParams = MappingProxyType(
        {
            "inputCols": ["image", "boxes"],
            "outputCol": "cropped_image",
            "keepInputData": False,
            "imageType": ImageType.FILE,
            "numPartitions": 0,
            "padding": 0,
            "pageCol": "page",
            "propagateError": False,
            "noCrop": True,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(ImageCropBoxes, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def transform_udf(self, image, data):
        if not isinstance(image, Image):
            image = Image(**image.asDict())
        try:
            if image.exception != "":
                return Image(
                    path=image.path,
                    imageType=image.imageType,
                    data=bytes(),
                    exception=image.exception,
                )
            img = image.to_pil()
            results = []
            for b in data.bboxes:
                box = b
                if not isinstance(box, Box):
                    box = Box(**box.asDict())
                if box.width > box.height:
                    results.append(
                        img.crop(box.bbox(self.getPadding())).rotate(-90, expand=True),
                    )
                else:
                    results.append(img.crop(box.bbox(self.getPadding())))
            if self.getNoCrop() and len(results) == 0:
                raise ImageCropError("No boxes to crop")
            if len(results) == 0:
                results.append(img)

        except Exception as e:
            exception = traceback.format_exc()
            exception = f"ImageCropBoxes: {exception}, {image.exception}"
            logging.warning(exception)
            if self.getPropagateError():
                raise ImageCropError from e
            return Image(image.path, image.imageType, data=bytes(), exception=exception)
        return Image.from_pil(results[0], image.path, image.imageType, image.resolution)

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        image_col = self._validate(self.getInputCols()[0], dataset)
        box_col = self._validate(self.getInputCols()[1], dataset)

        if self.getNumPartitions() > 0:
            dataset = dataset.repartition(self.getPageCol()).coalesce(
                self.getNumPartitions(),
            )
        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, Image.get_schema())(image_col, box_col),
        )

        if not self.getKeepInputData():
            result = result.drop(image_col)
        return result



import json
from types import MappingProxyType
from typing import Any

import pandas as pd
from pyspark import keyword_only

from scaledp.models.ner.BaseNer import BaseNer
from scaledp.schemas.Entity import Entity
from scaledp.schemas.NerOutput import NerOutput

from ...enums import Device


class Ner(BaseNer):

    defaultParams = MappingProxyType(
        {
            "inputCol": "text",
            "outputCol": "ner",
            "keepInputData": True,
            "model": "d4data/biomedical-ner-all",
            "whiteList": [],
            "numPartitions": 1,
            "partitionMap": False,
            "device": Device.CPU,
            "batchSize": 1,
            "scoreThreshold": 0.0,
            "pageCol": "page",
            "pathCol": "path",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(Ner, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    @staticmethod
    def split_text(text, max_length=500, stride=256):
        chunks = []
        for i in range(0, len(text), stride):
            chunk = text[i : i + max_length]
            chunks.append(chunk)
            if len(chunk) < max_length:
                break
        return chunks

    @staticmethod
    def aggregate_ner_results(text, pipeline, max_length=500, stride=256):
        chunks = Ner.split_text(text, max_length=max_length, stride=stride)
        entities = []
        for num, chunk in enumerate(chunks):
            ner_results = pipeline(chunk)
            new_ner_results = []
            for i in ner_results:
                i["start"] = i["start"] + num * stride
                i["end"] = i["end"] + num * stride
                new_ner_results.append(i)
            entities.extend(ner_results)
        return entities

    def get_pipeline(self):
        if self.pipeline is None:
            from transformers import (
                AutoModelForTokenClassification,
                AutoTokenizer,
                pipeline,
            )

            tokenizer = AutoTokenizer.from_pretrained(self.getModel())
            ort_model = AutoModelForTokenClassification.from_pretrained(self.getModel())
            self.pipeline = pipeline(
                "ner",
                model=ort_model,
                tokenizer=tokenizer,
                aggregation_strategy="first",
                device=int(self.getDevice()),
            )
        return self.pipeline

    def transform_udf(self, text):

        mapping = []
        for idx, box in enumerate(text.bboxes):
            mapping.extend([idx] * (len(box.text) + 1))

        result = Ner.aggregate_ner_results(
            text.text,
            self.get_pipeline(),
            max_length=500,
            stride=480,
        )

        entities = []
        for tag in result:
            if (
                len(self.getWhiteList()) > 0
                and tag["entity_group"] not in self.getWhiteList()
            ):
                continue
            boxes = mapping[tag["start"] : tag["end"]]
            boxes = [text.bboxes[i] for i in list(dict.fromkeys(boxes))]
            t = Entity(
                entity_group=tag["entity_group"],
                score=float(tag["score"]),
                word=tag["word"],
                start=tag["start"],
                end=tag["end"],
                boxes=boxes,
            )
            entities.append(t)
        return NerOutput(path=text.path, entities=entities, exception="")

    @staticmethod
    def transform_udf_pandas(
        texts: pd.DataFrame,
        params: pd.Series,
    ) -> pd.DataFrame:  # pragma: no cover
        params = json.loads(params[0])
        model = params["model"]
        from transformers import (
            AutoModelForTokenClassification,
            AutoTokenizer,
            pipeline,
        )

        tokenizer = AutoTokenizer.from_pretrained(model)
        model = AutoModelForTokenClassification.from_pretrained(model)
        pipe = pipeline(
            "ner",
            model=model,
            tokenizer=tokenizer,
            aggregation_strategy="first",
            device=int(params["device"]),
        )

        results = []
        max_length = 500
        stride = 480
        batch_texts = [
            (index, Ner.split_text(text["text"] or "-", max_length, stride))
            for index, text in texts.iterrows()
        ]

        if all(all(chunk == "-" for chunk in chunks) for index, chunks in batch_texts):
            for _index, text in texts.iterrows():
                output = NerOutput(path=text.path, entities=[], exception="")
                results.append(output)
            return pd.DataFrame(results)

        chunks = []
        offsets = []
        for index, exit_chunks in batch_texts:
            for num, chunk in enumerate(exit_chunks):
                chunks.append(chunk)
                offsets.append((index, num * stride))

        ner_results = pipe(chunks)
        batch_results = [[] for i in range(len(batch_texts))]
        for num, ner_result in enumerate(ner_results):

            for i in ner_result:
                i["start"] = i["start"] + offsets[num][1]
                i["end"] = i["end"] + offsets[num][1]
                batch_results[offsets[num][0]].append(i)

        for (_index, text), result in zip(texts.iterrows(), batch_results):
            mapping = []
            for idx, box in enumerate(text["bboxes"]):
                mapping.extend([idx] * (len(box["text"]) + 1))

            entities = []
            for tag in result:
                if (
                    len(params["whiteList"]) > 0
                    and tag["entity_group"] not in params["whiteList"]
                ):
                    continue
                if (
                    params["scoreThreshold"] > 0
                    and tag["score"] < params["scoreThreshold"]
                ):
                    continue
                boxes = mapping[tag["start"] : tag["end"]]
                boxes = [text.bboxes[i] for i in list(dict.fromkeys(boxes))]
                t = Entity(
                    entity_group=tag["entity_group"],
                    score=float(tag["score"]),
                    word=tag["word"],
                    start=tag["start"],
                    end=tag["end"],
                    boxes=boxes,
                )
                entities.append(t)
            output = NerOutput(path=text.path, entities=entities, exception="")
            results.append(output)

        return pd.DataFrame(results)



from types import MappingProxyType
from typing import Any

from pyspark import keyword_only

from scaledp.models.recognizers.BaseOcr import BaseOcr
from scaledp.params import (
    CODE_TO_LANGUAGE,
    LANGUAGE_TO_TESSERACT_CODE,
    Param,
    Params,
    TypeConverters,
)
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document

from ...enums import OEM, PSM, TessLib


class TesseractOcr(BaseOcr):
    """
    Run Tesseract OCR text recognition on images.
    """

    psm = Param(
        Params._dummy(),
        "psm",
        "The desired PageSegMode. Defaults to :attr:`PSM.AUTO",
        typeConverter=TypeConverters.toInt,
    )

    oem = Param(
        Params._dummy(),
        "oem",
        "OCR engine mode. Defaults to :attr:`OEM.DEFAULT`.",
        typeConverter=TypeConverters.toInt,
    )

    tessDataPath = Param(
        Params._dummy(),
        "tessDataPath",
        "Path to tesseract data folder.",
        typeConverter=TypeConverters.toString,
    )

    tessLib = Param(
        Params._dummy(),
        "tessLib",
        "The desired Tesseract library to use. Defaults to :attr:`TESSEROCR`",
        typeConverter=TypeConverters.toInt,
    )

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "text",
            "bypassCol": "",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "psm": PSM.AUTO,
            "oem": OEM.DEFAULT,
            "lang": ["eng"],
            "lineTolerance": 0,
            "keepFormatting": False,
            "tessDataPath": "/usr/share/tesseract-ocr/5/tessdata/",
            "tessLib": TessLib.PYTESSERACT,
            "partitionMap": False,
            "propagateError": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(TesseractOcr, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_pytesseract(cls, images, params):
        import pytesseract

        results = []

        config = (
            f"--psm {params['psm']} --oem {params['oem']} -l {cls.getLangTess(params)}"
        )
        for image, image_path in images:
            res = pytesseract.image_to_data(
                image,
                output_type=pytesseract.Output.DATAFRAME,
                config=config,
            )
            res["conf"] = res["conf"] / 100

            if not params["keepFormatting"]:
                res.loc[res["level"] == 4, "conf"] = 1.0
                res["text"] = res["text"].fillna("\n")

            res = res[res["conf"] > params["scoreThreshold"]][
                ["text", "conf", "left", "top", "width", "height"]
            ].rename(columns={"conf": "score", "left": "x", "top": "y"})
            res = res[res["text"] != "\n"]
            boxes = res.apply(
                lambda x: Box(*x).to_string().scale(1 / params["scaleFactor"]),
                axis=1,
            ).values.tolist()
            if params["keepFormatting"]:
                text = TesseractOcr.box_to_formatted_text(
                    boxes,
                    params["lineTolerance"],
                )
            else:
                text = " ".join([str(w) for w in res["text"].values.tolist()])

            results.append(
                Document(path=image_path, text=text, type="text", bboxes=boxes),
            )
        return results

    @staticmethod
    def getLangTess(params):
        return "+".join(
            [
                LANGUAGE_TO_TESSERACT_CODE[CODE_TO_LANGUAGE[lang]]
                for lang in params["lang"]
            ],
        )

    @classmethod
    def call_tesserocr(cls, images, params):  # pragma: no cover
        from tesserocr import RIL, PyTessBaseAPI, iterate_level

        results = []

        with PyTessBaseAPI(
            path=params["tessDataPath"],
            psm=params["psm"],
            oem=params["oem"],
            lang=cls.getLangTess(params),
        ) as api:
            api.SetVariable("debug_file", "ocr.log")
            api.SetVariable("save_blob_choices", "T")

            for image, image_path in images:
                api.SetImage(image)

                api.Recognize()
                iterator = api.GetIterator()
                boxes = []
                texts = []

                level = RIL.WORD
                for r in iterate_level(iterator, level):
                    conf = r.Confidence(level) / 100
                    text = r.GetUTF8Text(level)
                    box = r.BoundingBox(level)
                    if conf > params["scoreThreshold"]:
                        boxes.append(
                            Box(
                                text,
                                conf,
                                box[0],
                                box[1],
                                abs(box[2] - box[0]),
                                abs(box[3] - box[1]),
                            ).scale(1 / params["scaleFactor"]),
                        )
                        texts.append(text)
                if params["keepFormatting"]:
                    text = TesseractOcr.box_to_formatted_text(
                        boxes,
                        params["lineTolerance"],
                    )
                else:
                    text = " ".join(texts)

                results.append(
                    Document(
                        path=image_path,
                        text=text,
                        bboxes=boxes,
                        type="text",
                        exception="",
                    ),
                )
        return results

    @classmethod
    def call_ocr(cls, images, params):
        if params["tessLib"] == TessLib.TESSEROCR.value:
            return cls.call_tesserocr(images, params)
        if params["tessLib"] == TessLib.PYTESSERACT.value:
            return cls.call_pytesseract(images, params)
        raise ValueError(f"Unknown Tesseract library: {params['tessLib']}")

    def setPsm(self, value):
        """
        Sets the value of :py:attr:`psm`.
        """
        return self._set(psm=value)

    def getPsm(self):
        """
        Sets the value of :py:attr:`psm`.
        """
        return self.getOrDefault(self.psm)

    def setOem(self, value):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self._set(oem=value)

    def getOem(self):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self.getOrDefault(self.oem)

    def setTessDataPath(self, value):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self._set(tessDataPath=value)

    def getTessDataPath(self):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self.getOrDefault(self.tessDataPath)

    def setTessLib(self, value):
        """
        Sets the value of :py:attr:`tessLib`.
        """
        return self._set(tessLib=value)

    def getTessLib(self):
        """
        Gets the value of :py:attr:`tessLib`.
        """
        return self.getOrDefault(self.tessLib)
